#include "stdafx.h"
#include "onipatch.h"
#include <cstdio>

typedef unsigned char	u8;
typedef unsigned short	u16;
typedef unsigned long	u32;

struct PatchEntry
{
	u32	offset;
	u32	length;
	u8	data[0x100];	// just don't make patches bigger than this kthx
};

enum VersionType
{
	kType_Patch = 0,
	kType_Protected,
	kType_AlreadyPatched,
	kType_Done
};

struct FileVersion
{
	u32			checksum;
	VersionType	type;
	PatchEntry	* patchData;
	char		* name;
};

PatchEntry	kPatchList_World[] =
{
	{	0x0000BDA8,	0x00000031,
		{	0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 
			0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x8B, 0x0D, 0x00, 0x06, 0x56, 0x00, 0x68, 0x84, 0x10, 
			0x53, 0x00, 0x8B, 0x91, 0x78, 0x05, 0x00, 0x00, 0x52, 0xE8, 0x0A, 0x37, 0x11, 0x00, 0x83, 0xC4, 
			0x20 } },
	{	0x000247F0,	0x00000001,
		{	0xC3 } },
	{	0x00024860,	0x00000001,
		{	0xC3 } },
	{	0x00031C40,	0x00000001,
		{	0xC3 } },
	{	0x0007A0D0,	0x00000001,
		{	0xC3 } },
	{	0xFFFFFFFF }
};

PatchEntry	kPatchList_GE[] =
{
	// return from log functions
	{	0x00024ED0,	1,	{	0xC3 } },	// output to debugger.txt (unchecked vsprintf)
	{	0x00024F40,	1,	{	0xC3 } },	// output to debugger.txt (unchecked vsprintf)
	{	0x00024FB0,	1,	{	0xC3 } },	// output to startup.txt (unchecked vsprintf)
	{	0x00025020,	1,	{	0xC3 } },	// output to one of the standard outputs (unchecked vsprintf)
	{	0xFFFFFFFF }
};

// text in spanish
// file header has "Cracked by KeopS"
PatchEntry	kPatchList_SPCracked[] =
{
	{	0x00024830,	1,	{	0xC3 } },	// output to debugger.txt (unchecked vsprintf)
	{	0x000248A0,	1,	{	0xC3 } },	// output to debugger.txt (unchecked vsprintf)
	{	0x00024910,	1,	{	0xC3 } },	// output to startup.txt (unchecked vsprintf)
	{	0x00024980,	1,	{	0xC3 } },	// output to one of the standard outputs (unchecked vsprintf)
	{	0xFFFFFFFF }
};

// looks almost exactly like the world 1.0 version
// locally called 'oni_unknown'
// note: see if this is newer or older than what I'm calling 'world 1.0'
// has the same function offsets as the spanish cracked version
// however the executables were linked at seperate times; other offsets are different
PatchEntry	kPatchList_World2[] =
{
	{	0x00024830,	1,	{	0xC3 } },	// output to debugger.txt (unchecked vsprintf)
	{	0x000248A0,	1,	{	0xC3 } },	// output to debugger.txt (unchecked vsprintf)
	{	0x00024910,	1,	{	0xC3 } },	// output to startup.txt (unchecked vsprintf)
	{	0x00024980,	1,	{	0xC3 } },	// output to one of the standard outputs (unchecked vsprintf)
	{	0xFFFFFFFF }
};

// another recompile of the spanish version
PatchEntry	kPatchList_Unk1[] =
{
	{	0x00024830,	1,	{	0xC3 } },	// output to debugger.txt (unchecked vsprintf)
	{	0x000248A0,	1,	{	0xC3 } },	// output to debugger.txt (unchecked vsprintf)
	{	0x00024910,	1,	{	0xC3 } },	// output to startup.txt (unchecked vsprintf)
	{	0x00024980,	1,	{	0xC3 } },	// output to one of the standard outputs (unchecked vsprintf)
	{	0xFFFFFFFF }
};

FileVersion	kVersionList[] =
{
	{	0x1e166706,	kType_Patch,			kPatchList_World,		"World version 1.0" },
	{	0xe26c5d1c,	kType_Patch,			kPatchList_World2,		"World alt version" },
	{	0x71c073f6,	kType_Patch,			kPatchList_GE,			"German version 1.0" },
	{	0xf70ef851,	kType_Protected,		NULL,					"UK protected version" },
	{	0x5BB197B2,	kType_Patch,			kPatchList_SPCracked,	"Spanish version 1.0 (cracked)" },
	{	0xa0f04f4d,	kType_AlreadyPatched,	NULL,					"World version 1.0 (patched)" },
	{	0xe2dc5d8c,	kType_AlreadyPatched,	NULL,					"World alt version (patched)" },
	{	0x72307466,	kType_AlreadyPatched,	NULL,					"German version 1.0 (patched)" },
	{	0x5C219822,	kType_AlreadyPatched,	NULL,					"Spanish version 1.0 (cracked, patched)" },
	{	0x3647255D,	kType_Patch,			kPatchList_Unk1,		"Unknown version 1.0" },
	{	0x36B725CD,	kType_AlreadyPatched,	NULL,					"Unknown version 1.0 (patched)" },
	{	0x00000000,	kType_Done,				NULL }
};

int APIENTRY _tWinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPTSTR    lpCmdLine,
                     int       nCmdShow)
{
	OPENFILENAME	openDialogDescriptor;
	char			filePath[8192] = { 0 };

	memset(&openDialogDescriptor, 0, sizeof(openDialogDescriptor));

	openDialogDescriptor.lStructSize =	sizeof(OPENFILENAME);
	openDialogDescriptor.hwndOwner =	NULL;
	openDialogDescriptor.lpstrFile =	filePath;
	openDialogDescriptor.nMaxFile =		sizeof(filePath);
	openDialogDescriptor.lpstrTitle =	"Locate original copy of Oni";
	openDialogDescriptor.Flags =		OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_NOCHANGEDIR | OFN_PATHMUSTEXIST;

	if(GetOpenFileName(&openDialogDescriptor))
	{
		HANDLE	in = CreateFile(openDialogDescriptor.lpstrFile, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if(in)
		{
			u32	fileLength;

			fileLength = GetFileSize(in, NULL);

			if(fileLength != INVALID_FILE_SIZE)
			{
				u8	* fileData = new u8[fileLength];
				if(fileData)
				{
					u32	bytesTransferred;

					ReadFile(in, fileData, fileLength, &bytesTransferred, NULL);
					CloseHandle(in);
					in = NULL;

					if(fileLength == bytesTransferred)
					{
						u32	checksum = 0;

						for(u32 i = 0; i < fileLength; i++)
						{
							checksum += fileData[i];
							checksum = (checksum << 15) | (checksum >> (32 - 15));
						}

						FileVersion	* version;

						for(version = kVersionList; version->type != kType_Done; version++)
							if(version->checksum == checksum)
								break;

						switch(version->type)
						{
							case kType_Done:
							{
								char	message[1024];
								
								sprintf(message, "Executable checksum (%08X) not recognized: you have an unsupported version of Oni. Try applying the World 1.0 patch? (note: this is likely to fail)", checksum);
								
								if(MessageBox(	NULL,
												message,
												"onipatch: unsupported version",
												MB_YESNO | MB_ICONWARNING | MB_DEFBUTTON2) == IDYES)
									version = &kVersionList[0];
								else
									version = NULL;
							}
							break;
							
							case kType_Protected:
								MessageBox(	NULL,
											"You have a copy-protected version of Oni that cannot be unofficially patched. Go bug Take-Two to release an official patch, and ask around on the OniChars or Oni Central forums for a pre-patched version.",
											"onipatch: unsupported version",
											MB_OK | MB_ICONERROR);
								version = NULL;
								break;

							case kType_AlreadyPatched:
								MessageBox(	NULL,
											"The patch has already been applied to this version. Any problems still occuring must be byproducts of an overactive imagination.",
											"onipatch: already patched",
											MB_OK | MB_ICONERROR);
								version = NULL;
								break;
							
							case kType_Patch:
								break;
						}
						
						if(version)
						{
							PatchEntry	* patch = version->patchData;
							
							for(patch; patch->offset != 0xFFFFFFFF; patch++)
							{
								if((patch->offset + patch->length) <= fileLength)
								{
									memcpy(&fileData[patch->offset], patch->data, patch->length);
								}
							}

							filePath[0] = 0;

							memset(&openDialogDescriptor, 0, sizeof(openDialogDescriptor));

							openDialogDescriptor.lStructSize =	sizeof(OPENFILENAME);
							openDialogDescriptor.hwndOwner =	NULL;
							openDialogDescriptor.lpstrFile =	filePath;
							openDialogDescriptor.nMaxFile =		sizeof(filePath);
							openDialogDescriptor.lpstrTitle =	"Save patched executable";
							openDialogDescriptor.Flags =		OFN_NOCHANGEDIR | OFN_PATHMUSTEXIST;

							if(GetSaveFileName(&openDialogDescriptor))
							{
								HANDLE	out = CreateFile(openDialogDescriptor.lpstrFile, GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);

								if(out)
								{
									WriteFile(out, fileData, fileLength, &bytesTransferred, NULL);

									if(fileLength == bytesTransferred)
									{
										MessageBox(NULL, "Patch applied.", "onipatch", MB_OK);
									}
									else
									{
										MessageBox(NULL, "Error writing output file.", "onipatch: error", MB_OK);
									}

									CloseHandle(out);
								}
								else if(GetLastError() == ERROR_ALREADY_EXISTS)
								{
									MessageBox(NULL, "Don't overwrite your original Oni executable; keep it as a backup.", "onipatch: error", MB_OK);
								}
								else
								{
									MessageBox(NULL, "Couldn't open output file. You may be attempting to overwrite your original Oni executable; keep it as a backup.", "onipatch: error", MB_OK);
								}
							}
						}
					}
					else
					{
						MessageBox(NULL, "Error reading input file.", "onipatch: error", MB_OK);
					}
				}
				else
				{
					MessageBox(NULL, "Out of memory.", "onipatch: error", MB_OK);
				}
			}
			else
			{
				MessageBox(NULL, "Error reading input file.", "onipatch: error", MB_OK);
			}

			if(in)
			{
				CloseHandle(in);
			}
		}
		else
		{
			MessageBox(NULL, "Couldn't open input file.", "onipatch: error", MB_OK);
		}
	}

	return 0;
}
